package ru.dbotthepony.mc.otm.worldgen.placement

import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import it.unimi.dsi.fastutil.objects.ObjectRBTreeSet
import net.minecraft.core.BlockPos
import net.minecraft.util.RandomSource
import net.minecraft.util.valueproviders.FloatProvider
import net.minecraft.util.valueproviders.IntProvider
import net.minecraft.world.level.levelgen.placement.PlacementContext
import net.minecraft.world.level.levelgen.placement.PlacementModifier
import net.minecraft.world.level.levelgen.placement.PlacementModifierType
import ru.dbotthepony.mc.otm.core.collect.Vec3iHashStrategy
import ru.dbotthepony.mc.otm.registry.data.MPlacementModifiers
import ru.dbotthepony.mc.otm.worldgen.EnhancedPlacementContext
import ru.dbotthepony.mc.otm.worldgen.PlacementPos
import java.util.stream.Stream

/**
 * Ellipsoid ("cloud") placement
 */
data class EllipsoidPlacement(
	/**
	 * X position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val x: FloatProvider,

	/**
	 * Z position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val z: FloatProvider,

	/**
	 * Y position within ellipsoid sampler, expected to be in range -1 .. 1
	 */
	val y: FloatProvider,

	/**
	 * Total amount of block positions to generate
	 */
	val count: IntProvider,

	/**
	 * Ellipsoid size sampler on X axis
	 */
	val xLength: FloatProvider,

	/**
	 * Ellipsoid size sampler on Z axis
	 */
	val zLength: FloatProvider,

	/**
	 * Ellipsoid size sampler on Y axis
	 */
	val yLength: FloatProvider,
) : PlacementModifier(), EnhancedPlacement {
	init {
		require(xLength.minValue >= 1f) { "Bad ellipsoid x minimal size: $xLength" }
		require(zLength.minValue >= 1f) { "Bad ellipsoid z minimal size: $zLength" }
		require(yLength.minValue >= 1f) { "Bad ellipsoid y minimal size: $yLength" }
	}

	private fun evaluate(random: RandomSource, position: BlockPos): ObjectRBTreeSet<BlockPos> {
		val count = count.sample(random)
		val results = ObjectRBTreeSet<BlockPos>(Vec3iHashStrategy)

		if (count <= 0)
			return results

		val xLength = xLength.sample(random)
		val zLength = zLength.sample(random)
		val yLength = yLength.sample(random)

		val xPow = xLength * xLength
		val zPow = zLength * zLength
		val yPow = yLength * yLength

		var iterations = 0
		val maxIterations = count * 10

		while (results.size < count && ++iterations < maxIterations) {
			val x = this.x.sample(random) * xLength
			val y = this.y.sample(random) * yLength
			val z = this.z.sample(random) * zLength

			val isValidPoint = (x * x) / xPow +
				(y * y) / yPow +
				(z * z) / zPow <= 1.0f

			if (isValidPoint) {
				results.add(BlockPos(x.toInt() + position.x, y.toInt() + position.y, z.toInt() + position.z))
			}
		}

		return results
	}

	override fun getPositions(
		context: PlacementContext,
		random: RandomSource,
		position: BlockPos
	): Stream<BlockPos> {
		return evaluate(random, position).stream()
	}

	override fun evaluate(context: EnhancedPlacementContext, positions: List<PlacementPos>): List<PlacementPos> {
		val result = ArrayList<PlacementPos>()
		val results = positions.map { (it, c) -> evaluate(context.random, it).map { PlacementPos(it, c) } }
		result.ensureCapacity(result.size + results.sumOf { it.size } + 1)
		results.forEach { result.addAll(it) }
		return result
	}

	override val type: EnhancedPlacement.Type<*>
		get() = Companion

	override fun type(): PlacementModifierType<*> {
		return MPlacementModifiers.ELLIPSOID_PLACEMENT
	}

	companion object : EnhancedPlacement.Type<EllipsoidPlacement> {
		val CODEC: MapCodec<EllipsoidPlacement> by lazy {
			RecordCodecBuilder.mapCodec {
				it.group(
					FloatProvider.CODEC.fieldOf("x").forGetter(EllipsoidPlacement::x),
					FloatProvider.CODEC.fieldOf("y").forGetter(EllipsoidPlacement::y),
					FloatProvider.CODEC.fieldOf("z").forGetter(EllipsoidPlacement::z),
					IntProvider.CODEC.fieldOf("count").forGetter(EllipsoidPlacement::count),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("x_size").forGetter(EllipsoidPlacement::xLength),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("z_size").forGetter(EllipsoidPlacement::zLength),
					FloatProvider.codec(1f, Float.MAX_VALUE).fieldOf("y_size").forGetter(EllipsoidPlacement::yLength),
				).apply(it, ::EllipsoidPlacement)
			}
		}

		override val codec: MapCodec<EllipsoidPlacement>
			get() = CODEC
	}
}
